/* 
 *	HT Editor
 *	textedit.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htclipboard.h"
#include "textedit.h"

#include <stdlib.h>
#include <string.h>

int text_editor_pos_compare(text_editor_pos *a, text_editor_pos *b)
{
	if (a->line==b->line) {
		return a->pofs-b->pofs;
	}
	return a->line-b->line;
}

/*
 *	CLASS ht_text_editor
 */

// FIXME: put into registry
#define TEXT_EDITOR_TABSIZE			5
#define TEXT_EDITOR_HIGHLIGHT_LINEEND	true

void ht_text_editor::init(bounds *b, bool own_t, ht_textfile *t, bool own_l, ht_syntax_lexer *l, UINT e)
{
	ht_view::init(b, VO_OWNBUFFER | VO_SELECTABLE | VO_RESIZE, "text editor");
	VIEW_DEBUG_NAME("ht_text_editor");

	growmode=GM_VDEFORM | GM_HDEFORM;

	textfile=t;
	own_textfile=own_t;

	lexer=l;
	own_lexer=l;

	edit_options=e;
	top_line=0;
	xofs=0;
	cursorx=0;
	cursory=0;
	sel_start.line=0;
	sel_start.pofs=0;
	sel_end.line=0;
	sel_end.pofs=0;

	rowcol_indicator=NULL;
}

void ht_text_editor::done()
{
	if (own_textfile) {
		textfile->done();
		delete textfile;
	}

	if (own_lexer) {
		lexer->done();
		delete lexer;
	}

	ht_view::done();
}

UINT ht_text_editor::char_vsize(char c, UINT x)
{
	if (c=='\t') return TEXT_EDITOR_TABSIZE-x%TEXT_EDITOR_TABSIZE;
	return 1;
}

void ht_text_editor::clipboard_copy_cmd()
{
	if (text_editor_pos_compare(&sel_start, &sel_end)<0) {
		FILEOFS s, e;
		if (textfile->convert_line2ofs(sel_start.line, sel_start.pofs, &s) &&
		textfile->convert_line2ofs(sel_end.line, sel_end.pofs, &e)) {
			char dsc[1024]; /* FIXME: possible buffer overflow */
			sprintf(dsc, "%s::%s", textfile->get_name(), desc);
			clipboard_copy(dsc, textfile, s, e-s);
		}
	}
}

void ht_text_editor::clipboard_paste_cmd()
{
	FILEOFS o;
	UINT px=physical_cursorx();
	UINT py=top_line+cursory;
	UINT ll=textfile->getlinelength(py);
// fill gaps with spaces
	while (px>ll) {
		textfile->insert_chars(py, ll, (void*)" ", 1);
		ll++;
		o++;
	}
	textfile->convert_line2ofs(py, px, &o);
	clipboard_paste(textfile, o);
	UINT s=clipboard_getsize();
	text_editor_pos start, end;
	textfile->convert_ofs2line(o, &start.line, &start.pofs);
	textfile->convert_ofs2line(o+s, &end.line, &end.pofs);
	select_set(&start, &end);
}

void ht_text_editor::clipboard_cut_cmd()
{
}

void ht_text_editor::concat_lines(UINT a)
{
	UINT b=a+1;
	if (textfile->has_line(a) && textfile->has_line(b)) {
		UINT alen=textfile->getlinelength(a);
		char *aline=(char*)malloc(alen+1);
		UINT alinelen;
		textfile->getline(a, 0, aline, alen+1, &alinelen, NULL);
	
		text_editor_pos ss, se;

		ss=sel_start;
		se=sel_end;

		insert_chars(b, 0, aline, alinelen);
	
		delete_lines(a, 1);

		if (b>ss.line) {
			if (b==se.line) {
				se.pofs+=alen;
				se.line--;
			} else if (b<se.line) {
				se.line--;
			}
		} else {
			if (b==ss.line) {
				ss.pofs+=alen;
			}
			ss.line--;
			se.line--;
		}

		sel_start=ss;
		sel_end=se;
		normalize_selection();
	}
}

UINT ht_text_editor::cursor_up(UINT n)
{
	if (cursory>n) cursory-=n; else {
		n=scroll_up(n-cursory);
		cursory=0;
	}
	notify_cursor_changed();
	return n;
}

UINT ht_text_editor::cursor_down(UINT n)
{
	UINT lmh=textfile->linecount()-top_line;
	if (lmh>(UINT)size.h) lmh=size.h;
	if (cursory+n>lmh-1) {
		n=scroll_down(cursory+n-(lmh-1));
		lmh=textfile->linecount()-top_line;
		if (lmh>(UINT)size.h) lmh=size.h;
		cursory=lmh-1;
	} else cursory+=n;
	notify_cursor_changed();
	return n;
}

UINT ht_text_editor::cursor_left(UINT n)
{
	UINT p;
	if (cursorx+xofs>n) p=cursorx+xofs-n; else p=0;
	cursor_vput(p);
	return 1;
}

UINT ht_text_editor::cursor_right(UINT n)
{
	cursor_vput(cursorx+xofs+n);
	return 1;
}

void ht_text_editor::cursor_home()
{
	cursor_vput(0);
}

void ht_text_editor::cursor_end()
{
	cursor_vput(getline_vlength(top_line+cursory));
}

void ht_text_editor::cursor_pput(UINT dx)
{
	UINT vx=0, px=0;
	int v=cursorx+xofs;
	char line[1024];
	char *linep=line;

	UINT linelen;
	if (!textfile->getline(top_line+cursory, 0, line, sizeof line, &linelen, NULL)) return;
	
	while (linelen--) {
		if (px==dx) {
			v=-1;
			break;
		}
		int k=char_vsize(*(linep++), vx);
		vx+=k;
		v-=k;
		if (v<0) break;
		px++;
	}
	while (v>0) {
		if (px==dx) break;
		px++;
		vx++;
		v--;
	}
	if (xofs>vx) {
		cursorx=0;
		xofs=vx;
	} else {
		cursorx=vx-xofs;
		if (cursorx>(UINT)size.w-1) {
			xofs=cursorx-(size.w-1);
			cursorx=size.w-1;
		}
	}
	notify_cursor_changed();
}

void ht_text_editor::cursor_vput(UINT vx)
{
	if ((vx>=xofs) && (vx<xofs+size.w)) {
		cursorx=vx-xofs;
	} else if (vx<xofs) {
		xofs=vx;
		cursorx=0;
	} else if (vx>(UINT)size.w-1) {
		xofs=vx-(size.w-1);
		cursorx=size.w-1;
	} else {
		xofs=0;
		cursorx=vx;
	}
	notify_cursor_changed();
}

void ht_text_editor::delete_chars(UINT line, UINT ofs, UINT count)
{
	text_editor_pos pos;
	pos.line=line;
	pos.pofs=ofs;
	if ((sel_end.line==pos.line) &&
	(text_editor_pos_compare(&pos, &sel_start)>=0) &&
	(text_editor_pos_compare(&pos, &sel_end)<0)) {
		sel_end.pofs-=count;
	} else if ((sel_start.line==pos.line) &&
	(text_editor_pos_compare(&pos, &sel_start)<0)) {
		sel_start.pofs-=count;
		if (sel_end.line==pos.line) sel_end.pofs-=count;
	}
	textfile->delete_chars(line, ofs, count);
}

void ht_text_editor::draw()
{
/*	timer_handle h=new_timer();
	start_timer(h);*/

	UINT tabsize=TEXT_EDITOR_TABSIZE;
//	vcp c=getcolor(palidx_windialog_body);
	vcp bgcolor=VCP(VC_LIGHT(VC_YELLOW), VC_BLUE);
	clear(bgcolor);
	char line[1024];
	text_editor_pos pos;
	pos.line=top_line;
	for (int y=0; y<size.h; y++) {
		lexer_state state;

		pos.pofs=0;

//FIXME:debug:		if (!textfile->getline((top_line+y)|0x80000000, line, sizeof line, &state)) break;
		UINT linelen;
		if (!textfile->getline(top_line+y, 0, line, sizeof line, &linelen, &state)) break;
		line[linelen]=0;

		UINT x=0;
		if (lexer) {
			char *linep=(char*)line;
			UINT toklen;
			lexer_token tok;
			bool start_of_line=true;
			bool theend=false;
blabla:
			text_pos p;
			p.line=pos.line;
			p.pofs=pos.pofs;
			while ((tok=lexer->gettoken(linep, &state, &p, &toklen, start_of_line, false))) {
				UINT k, i;
				UINT vtoklen=toklen;
				bool print=true;
				bool is_tab=((toklen==1) && (*linep=='\t'));
				vcp color=lexer->gettoken_color(tok);
				
				if (is_tab) vtoklen=tabsize-x%tabsize;
			
				if (x>=xofs) {
					k=x-xofs;
					i=0;
					if (k>(UINT)size.w-1) {
						theend=true;
						break;
					}
				} else if (x+toklen>=xofs) {
					k=0;
					i=xofs-x;
				} else {
					print=false;
				}
				if (print) {
					if (is_tab) {
						char tab[17];
						UINT z;
						for (z=0; z<vtoklen; z++) tab[z]=' ';
						tab[z]=0;
						render_str(k, y, color, &pos, vtoklen-i, tab+i, false);
					} else {
						render_str(k, y, color, &pos, vtoklen-i, linep+i, true);
					}
				}
				x+=vtoklen;
				linep+=toklen;
				linelen-=toklen;
				pos.pofs+=toklen;
				start_of_line=false;
			}
			if ((linelen) && (!*linep) && (!theend)) {
				while ((linelen) && (!*linep)) {
					UINT k, i;
					bool print=true;
					if (x>=xofs) {
						k=x-xofs;
						i=0;
						if (k>(UINT)size.w-1) {
							theend=true;
							break;
						}
					} else if (x+toklen>=xofs) {
						k=0;
						i=xofs-x;
					} else {
						print=false;
					}
					if (print) render_str(k, y, bgcolor, &pos, 1, " ", false);
					linep++;
					linelen--;
					pos.pofs++;
					x++;
				}
				if (linelen && !theend) goto blabla;
			}
		} else {
			char *linep=line;
			while (linelen--) {
				UINT vtoklen=char_vsize(*linep, x);
				if (x>=xofs) {
					if (x-xofs>(UINT)size.w-1) break;
					if (*linep=='\t') {
						vcp c=bgcolor;
						char tab[17];
						UINT z;
						for (z=0; z<vtoklen; z++) tab[z]=' ';
						tab[z]=0;
						render_str_color(&c, &pos);
						buf_lprint(x-xofs, y, c, z, tab);
					} else {
						vcp c=bgcolor;
						render_str_color(&c, &pos);
						buf_printchar(x-xofs, y, c, *linep);
					}
				}
				x+=vtoklen;
				linep++;
				pos.pofs++;
			}
		}
		if ((TEXT_EDITOR_HIGHLIGHT_LINEEND) && (pos.line<sel_end.line) &&
		(pos.line>=sel_start.line)) {
			fill(x, y, size.w-x, 1, VCP(VC_TRANSPARENT, VC_WHITE), ' ');
		}
		pos.line++;
	}
	if (focused) setcursor(cursorx, cursory);

/*	stop_timer(h);
	int tix=get_timer_tick(h);
	delete_timer(h);
	buf_printf(50, 0, bgcolor, "%dtix", tix);*/
}

UINT ht_text_editor::getline_vlength(UINT line)
{
	char l[1024];
	char *linep=l;
	UINT vl=0;

	UINT linelen;
	if (!textfile->getline(line, 0, l, sizeof l, &linelen, NULL)) return 0;
	
	while (linelen--) vl+=char_vsize(*(linep++), vl);
	
	return vl;
}

void ht_text_editor::handlemsg(htmsg *msg)
{
	switch (msg->msg) {
		case msg_keypressed: {
			int k=msg->data1.integer;
			bool shift;
			switch (k) {
/* FIXME: debug */
/*				case 'X':
					textfile->cntl(FCNTL_CACHE_INVD);
					dirtyview();
					clearmsg(msg);
					return;*/
				case K_Up:
				case K_Shift_Up:
					shift=(k==K_Shift_Up);
					if (shift) select_start();
					cursor_up(1);
					if (shift) select_end();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Down:
				case K_Shift_Down:
					shift=(k==K_Shift_Down);
					if (shift) select_start();
					cursor_down(1);
					if (shift) select_end();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Left:
				case K_Shift_Left:
					shift=(k==K_Shift_Left);
					if (shift) select_start();
					cursor_left(1);
					if (shift) select_end();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Right:
				case K_Shift_Right:
					shift=(k==K_Shift_Right);
					if (shift) select_start();
					cursor_right(1);
					if (shift) select_end();
					dirtyview();
					clearmsg(msg);
					return;
				case K_PageUp:
				case K_Shift_PageUp:
					shift=(k==K_Shift_PageUp);
					if (shift) select_start();
					scroll_up(size.h-1);
					if (shift) select_end();
					dirtyview();
					clearmsg(msg);
					return;
				case K_PageDown:
				case K_Shift_PageDown:
					shift=(k==K_Shift_PageDown);
					if (shift) select_start();
					scroll_down(size.h-1);
					if (shift) select_end();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Home:
				case K_Shift_Home:
					shift=(k==K_Shift_Home);
					if (shift) select_start();
					cursor_home();
					if (shift) select_end();
					dirtyview();
					clearmsg(msg);
					return;
				case K_End:
				case K_Shift_End:
					shift=(k==K_Shift_End);
					if (shift) select_start();
					cursor_end();
					if (shift) select_end();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Control_PageUp:
					top_line=0;
					cursorx=0;
					cursory=0;
					notify_cursor_changed();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Control_PageDown:
					top_line=textfile->linecount()-1;
					if (top_line>(UINT)size.h-1) top_line-=size.h-1;
					cursorx=0;
					cursory=0;
					scroll_down(size.h-1);
					notify_cursor_changed();
					dirtyview();
					clearmsg(msg);
					return;
				case K_Return:
					if (!(edit_options & TEXTEDITOPT_READONLY)) {
						split_line(top_line+cursory, physical_cursorx());
						cursor_home();
						cursor_down(1);
						dirtyview();
						clearmsg(msg);
						return;
					}						
					break;
				case K_Delete: {
					if (!(edit_options & TEXTEDITOPT_READONLY)) {
						UINT cx=physical_cursorx();
						if (cx<textfile->getlinelength(top_line+cursory)) {
							UINT px=physical_cursorx();
							delete_chars(top_line+cursory, cx, 1);
							cursor_pput(px);
						} else {
							concat_lines(top_line+cursory);
						}
						dirtyview();
						clearmsg(msg);
						return;
					}
					break;						
				}
				case K_BackSpace: {
					if (!(edit_options & TEXTEDITOPT_READONLY)) {
						UINT cx=physical_cursorx();
						if (cx) {
							UINT px=physical_cursorx()-1;
							delete_chars(top_line+cursory, cx-1, 1);
							cursor_pput(px);
						} else {
							if (top_line+cursory) {
								UINT la=top_line+cursory-1;
								cursor_up(1);
								cursor_end();
								concat_lines(la);
							}
						}
						dirtyview();
						clearmsg(msg);
						return;
					}
					break;						
				}
				case K_Control_Insert:
					sendmsg(cmd_clipboard_copy);
					clearmsg(msg);
					return;
				case K_Shift_Insert:
					sendmsg(cmd_clipboard_paste);
					clearmsg(msg);
					return;
				case K_Shift_Delete:
					sendmsg(cmd_clipboard_cut);
					clearmsg(msg);
					return;
				default: {
					if (!(edit_options & TEXTEDITOPT_READONLY)) {
						int k=msg->data1.integer;
						if (((k>=' ') && (k<=255)) ||
						((k=='\t') && (edit_options & TEXTEDITOPT_INPUTTABS))) {
							char s[2];
							s[0]=k;
							s[1]=0;
							UINT px=physical_cursorx();
							insert_chars(top_line+cursory, px, s, 1);
							cursor_pput(px);
							cursor_right(char_vsize(k, cursorx+xofs));
							dirtyview();
							clearmsg(msg);
							return;
						}							
					}
					break;
				}
			}
			break;
		}
		case cmd_clipboard_copy: {
			clipboard_copy_cmd();
			dirtyview();
			clearmsg(msg);
			return;
		}
		case cmd_clipboard_paste: {
			if (!(edit_options & TEXTEDITOPT_READONLY)) {
				clipboard_paste_cmd();
				dirtyview();
				clearmsg(msg);
				return;
			}
			break;				
		}
		case cmd_clipboard_cut: {
			if (!(edit_options & TEXTEDITOPT_READONLY)) {
				clipboard_cut_cmd();
				dirtyview();
				clearmsg(msg);
				return;
			}
			break;				
		}
	}
	ht_view::handlemsg(msg);
}

void ht_text_editor::insert_chars(UINT line, UINT ofs, void *chars, UINT len)
{
	text_editor_pos pos;
	pos.line=line;
	pos.pofs=ofs;
	if ((sel_end.line==pos.line) &&
	(text_editor_pos_compare(&pos, &sel_start)>=0) &&
	(text_editor_pos_compare(&pos, &sel_end)<0)) {
		sel_end.pofs+=len;
	} else if ((sel_start.line==pos.line) &&
	(text_editor_pos_compare(&pos, &sel_start)<0)) {
		sel_start.pofs+=len;
		if (sel_end.line==pos.line) sel_end.pofs+=len;
	}
	textfile->insert_chars(line, ofs, chars, len);
}

void ht_text_editor::make_pos_physical(text_editor_pos *p)
{
	UINT l=textfile->getlinelength(p->line);
	if (p->pofs > l) p->pofs=l;
}

void ht_text_editor::notify_cursor_changed()
{
	if (rowcol_indicator) {
		ht_rowcol_indicator_data d;
		d.row=top_line+cursory+1;
		d.col=xofs+cursorx+1;
		rowcol_indicator->databuf_set(&d);
	}
}

void ht_text_editor::normalize_selection()
{
	if ((text_editor_pos_compare(&sel_end, &sel_start)<=0)) {
		sel_start.line=0;
		sel_start.pofs=0;
		sel_end.line=0;
		sel_end.pofs=0;
	}
}

UINT ht_text_editor::physical_cursorx()
{
	int vx=0, px=0, v=cursorx+xofs;
	char line[1024];
	char *linep=line;

	UINT linelen;
	if (!textfile->getline(top_line+cursory, 0, line, sizeof line, &linelen, NULL)) return 0;
	
	while (linelen--) {
		int k=char_vsize(*(linep++), vx);
		vx+=k;
		v-=k;
		if (v<0) break;
		px++;
	}
	if (v>0) px+=v;
	
	return px;
}

void ht_text_editor::render_str(int x, int y, vcp color, text_editor_pos *pos, UINT len, char *str, bool multi)
{
	if ((pos->line == sel_start.line) || (pos->line == sel_end.line)) {
		text_editor_pos p=*pos;
		vcp c;
		while (len--) {
			c=color;
			render_str_color(&c, &p);
			buf_lprint(x++, y, c, 1, str++);
			if (multi) p.pofs++;
		}
	} else {
		render_str_color(&color, pos);
		buf_lprint(x, y, color, len, str);
	}
}

void ht_text_editor::render_str_color(vcp *color, text_editor_pos *pos)
{
	if ((text_editor_pos_compare(pos, &sel_start)>=0) &&
	(text_editor_pos_compare(pos, &sel_end)<0)) {
		if (VCP_FOREGROUND(*color)==VC_WHITE) {
			*color=VCP(VC_LIGHT(VC_BLACK), VC_WHITE);
		} else {
			*color=vcp_mix(*color, VCP(VC_TRANSPARENT, VC_WHITE));
		}
	}
}

void ht_text_editor::resize(int rw, int rh)
{
	ht_view::resize(rw, rh);
	if ((int)cursorx>size.w-1) {
		xofs+=(int)cursorx-(size.w-1);
		cursorx=size.w-1;
	}
	if ((int)cursory>size.h-1) {
		top_line+=(int)cursory-(size.h-1);
		cursory=size.h-1;
	}
	notify_cursor_changed();
}

UINT ht_text_editor::scroll_up(UINT n)
{
	if (top_line>n) top_line-=n; else {
		int q=n-top_line;
		top_line=0;
		cursory=0;
		notify_cursor_changed();
		return q;
	}
	notify_cursor_changed();
	return n;
}

UINT ht_text_editor::scroll_down(UINT n)
{
	UINT lc=textfile->linecount();
	if (top_line+n+size.h < lc) top_line+=n; else {
		if (lc-top_line>=(UINT)size.h) {
			int q=top_line;
			top_line=lc-size.h;
			cursory=size.h-1;
			notify_cursor_changed();
			return top_line-q;
		}
		cursory=lc-1;
		notify_cursor_changed();
		return 0;
	}
	notify_cursor_changed();
	return n;
}

UINT ht_text_editor::scroll_left(UINT n)
{
	if (xofs>n) xofs-=n; else xofs=0;
	notify_cursor_changed();
	return n;
}

UINT ht_text_editor::scroll_right(UINT n)
{
	xofs+=n;
	notify_cursor_changed();
	return n;
}

void ht_text_editor::select_add(text_editor_pos *s, text_editor_pos *e)
{
	text_editor_pos start=*s, end=*e;
	make_pos_physical(&start);
	make_pos_physical(&end);
	if (text_editor_pos_compare(&start, &end)>0) {
		text_editor_pos temp=start;
		start=end;
		end=temp;
	}
	
	if ((text_editor_pos_compare(&end, &sel_start)==0)) {
		sel_start=start;
	} else if (text_editor_pos_compare(&start, &sel_end)==0) {
		sel_end=end;
	} else if (text_editor_pos_compare(&start, &sel_start)==0) {
		sel_start=end;
	} else if (text_editor_pos_compare(&end, &sel_end)==0) {
		sel_end=start;
	} else {
		sel_start=start;
		sel_end=end;
	}
	
	if (text_editor_pos_compare(&sel_start, &sel_end)>0) {
		text_editor_pos temp=sel_start;
		sel_start=sel_end;
		sel_end=temp;
	}
	normalize_selection();
}

void ht_text_editor::select_set(text_editor_pos *s, text_editor_pos *e)
{
	sel_start=*s;
	sel_end=*e;
	normalize_selection();
}

void ht_text_editor::select_start()
{
	selectmode=true;
	selectstart.line=top_line+cursory;
	selectstart.pofs=physical_cursorx();
}

void ht_text_editor::select_end()
{
	text_editor_pos p;
	selectmode=false;
	p.line=top_line+cursory;
	p.pofs=physical_cursorx();
	select_add(&selectstart, &p);
}

void ht_text_editor::set_rowcol_indicator(ht_rowcol_indicator *r)
{
	rowcol_indicator=r;
	notify_cursor_changed();
}

void ht_text_editor::delete_lines(UINT line, UINT count)
{
	if (sel_start.line+1>line) sel_start.line--;
	if (sel_end.line>line) sel_end.line--;
	normalize_selection();
	textfile->delete_lines(line, count);
}

void ht_text_editor::insert_lines(UINT line, UINT count)
{
	if (sel_start.line+1>line) sel_start.line++;
	if (sel_end.line>line) sel_end.line++;
	normalize_selection();
	textfile->insert_lines(line, count);
}

void ht_text_editor::split_line(UINT a, UINT pos)
{
	UINT l=textfile->getlinelength(a);
	if (pos>l) pos=l;
	char *aline=(char*)malloc(pos+1);
	UINT alinelen;
	textfile->getline(a, 0, aline, pos+1, &alinelen, NULL);

	text_editor_pos p, ss, se;
	p.line=a;
	p.pofs=pos;

	ss=sel_start;
	se=sel_end;

	insert_lines(a, 1);
	insert_chars(a, 0, aline, alinelen);

	delete_chars(a+1, 0, pos);
	free(aline);
	
	if (text_editor_pos_compare(&p, &ss)>0) {
		if (text_editor_pos_compare(&p, &se)<0) {
			if (se.line==p.line) {
				se.pofs-=pos;
			}
			se.line++;
		}
	} else {
		if (ss.line==p.line) {
			ss.pofs-=pos;
		}
		ss.line++;
		se.line++;
	}
	sel_start=ss;
	sel_end=se;
	normalize_selection();
}

/*
 *	CLASS ht_rowcol_indicator
 */
 
void ht_rowcol_indicator::init(bounds *b)
{
	ht_view::init(b, VO_RESIZE, NULL);
	row=0;
	col=0;

	growmode=GM_LEFT | GM_BOTTOM;
}

void ht_rowcol_indicator::draw()
{
	vcp bgcolor=getcolor(palidx_windialog_body);
//     clear(bgcolor);
	char buf[32];
	sprintf(buf, " %d:%d ", row, col);
	int l=strlen(buf);
	if (l>size.w) l=size.w;
	buf_print((size.w-l)/2, 0, bgcolor, buf);
}

void ht_rowcol_indicator::getdata(ht_object_stream *s)
{
	s->put_int_dec(row, 4, NULL);
	s->put_int_dec(col, 4, NULL);
}

void ht_rowcol_indicator::setdata(ht_object_stream *s)
{
	row=s->get_int_dec(4, NULL);
	col=s->get_int_dec(4, NULL);
	dirtyview();
}

